import numpy as np

class ActivationModel:
    def activation(
        self,
        Z
    ):
        # Z: (num_outputs, num_examples)
        # outputs: (num_outputs, num_examples)
        raise NotImplementedError('Method must be implemented by child class')

    def d_activation_d_Z(
        self,
        Z,
        outputs
    ):
        # Z: (num_outputs, num_examples)
        # outputs: (num_outputs, num_examples)
        # d_activation_d_Z: (num_outputs, num_examples)
        raise NotImplementedError('Method must be implemented by child class')

class SigmoidActivationModel(ActivationModel):
    def activation(
        self,
        Z
    ):
        # Z: (num_outputs, num_examples)
        # outputs: (num_outputs, num_examples)
        return np.reciprocal(np.add(1.0, np.exp(-Z))) # (num_outputs, num_examples)

    def d_activation_d_Z(
        self,
        Z,
        outputs
    ):
        # Z: (num_outputs, num_examples)
        # outputs: (num_outputs, num_examples)
        # d_activation_d_Z: (num_outputs, num_examples)
        return np.multiply(outputs, np.subtract(1, outputs)) # (num_outputs, num_examples)

class TanhActivationModel(ActivationModel):
    def activation(
        self,
        Z
    ):
        # Z: (num_outputs, num_examples)
        # outputs: (num_outputs, num_examples)
        return np.tanh(Z) # (num_outputs, num_examples)

    def d_activation_d_Z(
        self,
        Z,
        outputs
    ):
        # Z: (num_outputs, num_examples)
        # outputs: (num_outputs, num_examples)
        # d_activation_d_Z: (num_outputs, num_examples)
        return np.subtract(1, np.square(outputs)) # (num_outputs, num_examples)

class ReLUActivationModel(ActivationModel):
    def activation(
        self,
        Z
    ):
        # Z: (num_outputs, num_examples)
        # outputs: (num_outputs, num_examples)
        return np.multiply(Z, Z > 0.0) # (num_outputs, num_examples)

    def d_activation_d_Z(
        self,
        Z,
        outputs
    ):
        # Z: (num_outputs, num_examples)
        # outputs: (num_outputs, num_examples)
        # d_activation_d_Z: (num_outputs, num_examples)
        return np.multiply(1.0, Z > 0.0) # (num_outputs, num_examples)



from tqdm.autonotebook import tqdm

class NeuralNet:
    def __init__(
        self,
        layers,
        cost_model,
        prediction_model
    ):
        self.layers = layers
        self.cost_model = cost_model
        self.prediction_model = prediction_model
        self.num_layers = len(layers)

    def train(
        self,
        X,
        y,
        num_iterations,
        learning_rate,
        progress_bar=False
    ):
        if progress_bar:
            training_iterator = tqdm(range(num_iterations))
        else:
            training_iterator = range(num_iterations)
        for iteration_index in training_iterator:
            inputs = X
            for layer in self.layers:
                outputs = layer.forward_propagate(inputs)
                inputs = outputs
            cost = self.cost_model.cost(outputs, y)
            d_cost_d_outputs = self.cost_model.d_cost_d_outputs(outputs, y)
            for layer in reversed(self.layers):
                d_cost_d_inputs = layer.backward_propagate(d_cost_d_outputs)
                d_cost_d_outputs = d_cost_d_inputs
            for layer in self.layers:
                layer.weights += -learning_rate*layer.d_cost_d_weights
                layer.biases += -learning_rate*layer.d_cost_d_biases
            # if progress_bar:
            #     training_iterator.set_postfix(cost=cost)
            if iteration_index % 100 == 0:
                if progress_bar:
                    training_iterator.set_postfix(cost=cost)
                else:
                    print('Iteration {}: cost {}'.format(
                        iteration_index,
                        cost
                    ))

    def predict(
        self,
        X
    ):
        inputs = X
        for layer in self.layers:
            outputs = layer.forward_propagate(inputs)
            inputs = outputs
        y_hat = self.prediction_model.predictions(outputs)
        return y_hat



import numpy as np

class Layer:
    def __init__(
        self,
        num_inputs,
        num_outputs,
        activation_model,
        weight_initialization_multiplier = 0.01,
        bias_initialization_multiplier = 0.0
    ):
        # weights: (num_outputs, num_inputs)
        # biases: (num_outputs, 1)
        self.num_inputs = num_inputs
        self.num_outputs = num_outputs
        self.activation_model = activation_model
        self.weight_initialization_multiplier = weight_initialization_multiplier
        self.bias_initialization_multiplier = bias_initialization_multiplier
        self.weights = np.multiply(
            np.random.randn(self.num_outputs, self.num_inputs),
            self.weight_initialization_multiplier
        ) # (num_outputs, num_inputs)
        self.biases = np.multiply(
            np.random.randn(self.num_outputs, 1),
            self.bias_initialization_multiplier
        ) # (num_outputs, 1)

    def forward_propagate(
        self,
        inputs
    ):
        # inputs: (num_inputs, num_examples)
        # Z: (num_outputs, num_examples)
        # outputs: (num_outputs, num_examples)
        self.inputs = inputs # (num_inputs, num_examples)
        self.Z = np.dot(self.weights, self.inputs) + self.biases # (num_outputs, num_examples)
        self.outputs = self.activation_model.activation(self.Z) # (num_outputs, num_examples)
        return self.outputs

    def backward_propagate(
        self,
        d_cost_d_outputs
    ):
        # d_cost_d_outputs: (num_outputs, num_examples)
        # d_cost_d_Z: (num_outputs, num_examples)
        # d_cost_d_weights (num_outputs, num_inputs)
        # d_cost_d_biases: (num_outputs, 1)
        # d_cost_d_inputs: (num_inputs, num_examples)
        self.d_cost_d_outputs = d_cost_d_outputs # (num_outputs, num_examples)
        self.d_cost_d_Z = self.d_cost_d_outputs * self.activation_model.d_activation_d_Z(self.Z, self.outputs) # (num_outputs, num_examples)
        self.d_cost_d_weights = np.dot(self.d_cost_d_Z, self.inputs.T) # (num_outputs, num_inputs)
        self.d_cost_d_biases = np.sum(self.d_cost_d_Z, axis=1, keepdims=True) # (num_outputs, 1)
        self.d_cost_d_inputs = np.dot(self.weights.T, self.d_cost_d_Z) # (num_inputs, num_examples)
        return self.d_cost_d_inputs



import numpy as np

class CostModel:
    def cost(outputs, y):
        # outputs: (num_outputs, num_examples)
        # cost: Scalar
        raise NotImplementedError('Method must be implemented by child class')

    def d_cost_d_outputs(outputs, y):
        # outputs: (num_outputs, num_examples)
        # d_cost_d_outputs: (num_outputs, num_examples)
        raise NotImplementedError('Method must be implemented by child class')

class CrossEntropyBinaryClassificationCostModel(CostModel):
    def cost(
        self,
        outputs,
        y
    ):
        # outputs: (1, num_examples)
        # y: (1, num_examples)
        # cost: Scalar

        # outputs: For this cost function, outputs[0, i] is interpreted as the
        # probability that the ith example is of class 1 (as opposed to class 0)

        # y: For this cost function, y[0, i] is either 0
        # (first class) or 1 (second class)

        num_examples = y.shape[1]
        return np.squeeze(-(1/num_examples)*np.sum(
            np.add(
                np.multiply(
                    y,
                    np.log(outputs)
                ),
                np.multiply(
                    np.subtract(1.0, y),
                    np.log(np.subtract(1.0, outputs))
                )
            ),
            axis=1
        ))

    def d_cost_d_outputs(
        self,
        outputs,
        y
    ):
        # outputs: (num_outputs, num_examples)
        # y: (num_classes, num_examples)
        # d_cost_d_outputs: (num_outputs, num_examples)

        # outputs: For this cost function, outputs[0, i] is interpreted as the
        # probability that the ith example is of class 1 (as opposed to class 0)

        # y: For this cost function, y[0, i] is either 0
        # (first class) or 1 (second class)

        num_examples = y.shape[1]
        return -(1/num_examples)*np.divide(
            np.subtract(
                y,
                outputs
            ),
            np.multiply(
                outputs,
                np.subtract(1, outputs)
            )
        )



import numpy as np

class PredictionModel:
    def predictions(
        self,
        outputs
    ):
        # outputs: (num_outputs, num_examples)
        # predictions: (?, num_examples)
        raise NotImplementedError('Method must be implemented by child class')

class BinaryClassificationPredictionModel(PredictionModel):
    def predictions(
        self,
        outputs
    ):
        # outputs: (1, num_examples)
        # predictions: (1, num_examples)
        y_hat = np.around(outputs).astype('int')
        return y_hat



import os
from setuptools import setup, find_packages

BASEDIR = os.path.dirname(os.path.abspath(__file__))
VERSION = open(os.path.join(BASEDIR, 'VERSION')).read().strip()

# Dependencies (format is 'PYPI_PACKAGE_NAME[>=]=VERSION_NUMBER')
BASE_DEPENDENCIES = [
    'numpy>=1.18',
    'tqdm>=4.46'
]

# TEST_DEPENDENCIES = [
# ]
#
# LOCAL_DEPENDENCIES = [
# ]

# Allow setup.py to be run from any path
os.chdir(os.path.normpath(BASEDIR))

setup(
    name='simple-neural-net',
    packages=find_packages(),
    version=VERSION,
    include_package_data=True,
    description='Minimal framework for creating a simple neural net',
    long_description=open('README.md').read(),
    url='https://github.com/tcquinn/simple-neural-net',
    author='Theodore Quinn',
    author_email='tcquinn@tcquinn.org',
    install_requires=BASE_DEPENDENCIES,
    # tests_require=TEST_DEPENDENCIES,
    # extras_require = {
    #     'test': TEST_DEPENDENCIES,
    #     'local': LOCAL_DEPENDENCIES
    # },
    keywords=['neural network', 'deep learning'],
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ]
)



from simple_neural_net.core import *
from simple_neural_net.layers import *
from simple_neural_net.activation_models import *
from simple_neural_net.cost_models import *
from simple_neural_net.prediction_models import *
